// This file is released under the 3-clause BSD license. See COPYING-BSD.
// The algorithm used here is described in the book "La commande predictive" 
// from Patrick Boucher and Didier Dumur at chapter 3 page 51. 
// ISBN : 9782710807025 (176 pages) edited in 1996.


function coeffpzmoins1 = convertztozmoins1(pz, degree_pdiv)
    coeffpzmoins1 = coeff(pz);

    for (i = degree(pz) + 2:degree_pdiv + 1)
        coeffpzmoins1(i) = 0;
    end
    coeffpzmoins1 = flipdim(coeffpzmoins1, 2);
endfunction

function fij = computefij (i, j)
    if (j == 1) then
        fij = - alphai(i+2);
    else 
        if (i == na) then
            fij = -1.0 * computefij(0, j-1) * alphai(na + 2);
        else
            fij = computefij(i+1, j-1) - computefij(0, j-1) * alphai(i+2);
        end
    end
endfunction

function Fj = computeFj (j)
    Fj = 0;
    for (i = 0:na)
        fij = computefij (i, j);
        Fj = Fj + fij * z^-i;
    end
endfunction

function hij = computehij (i, j)
    if (j == 1) then
        hij = Bi(i + 3);
    else
        if (i == nb - 2) then
            hij = computefij(0, j-1) * Bi(nb + 1);
        else
            hij = computehij(i+1, j-1) + computefij(0, j-1) * Bi(i + 3);
        end
    end
endfunction

function Hj = computeHj (j)
    Hj = 0;
    for (i = 0:nb - 2)
        hij = computehij(i, j);
        Hj = Hj + hij * z^-i;
    end
endfunction

function [R, Sfinal, T, Hboc] = compute_predictive_rst(linear_sys, Te, Tfstep, N1, N2, Nu, Lambda)
    // discretization with Te sampling
    discrete_sys = dscr(linear_sys, Te);
    discrete_tf = ss2tf(discrete_sys);

    // step response for modeling
    numberofTe = Tfstep / Te;
    ustep = ones(1, numberofTe);
    s1 = dsimul(discrete_sys, ustep);

    // prepare model
    Bz = discrete_tf.num;
    Az = discrete_tf.den;
    Ai = convertztozmoins1(Az, degree(Az));
    Bi = convertztozmoins1(Bz, degree(Az));
    nb = size(Bi,2) - 1;
    na = size(Ai,2) - 1;

    // G computation
    for (i = 1:(N2 - N1 + 1))
        for (j = 1:Nu)
            indice_s1 = N1 + i - j + 1;
            if (indice_s1 < 1) then
                G(i, j) = 0.0;
            else
                G(i,j) = s1(indice_s1);
            end
        end
    end

    z = poly(0, "z");

    // computation of alpha polynome
    alpha = 0;
    for (i = 0:na + 1)
        if (i == 0) then
            alphai(i + 1) = 1;
        else
            if (i == na + 1) then
                alphai(i + 1) = -Ai(na + 1);
            else
                alphai(i + 1) = Ai(i+1) - Ai(i);
            end
        end
        alpha = alpha + alphai(i+1) * z^-i;
    end

    // M computation
    M = inv(G' * G + Lambda * eye(Nu, Nu)) * G';
    M = M(1,:);

    // if computation
    ifm = [];
    for (j = 1:N2 - N1 + 1)
        ifm = [ifm; computeFj(N1 + j - 1)];
    end

    // ih computation
    ihm = [];
    for (j = 1:N2 - N1 + 1)
        ihm = [ihm; computeHj(N1 + j - 1)];
    end

    // R(q^-1) S(q^-1) T(q) computation
    S = 1 + M * ihm * z^-1;
    R = M * ifm;
    wtemp = [];
    for (i = 1:N2 - N1 + 1)
        wtemp = [wtemp; z^(N1 + i - 1)];
    end
    T = M * wtemp;
    Sfinal = ((1 - z^-1) * S)^-1;

    // stability analysis
    Hboc = Sfinal * T * discrete_tf / (1 + R * Sfinal * discrete_tf);
    racines = roots(Hboc.den);
    gain_margin = g_margin(Hboc);
    phase_margin = p_margin(Hboc);
    printf("\n");
    printf("For stability purpose :\n");
    printf("- gain margin should be inside [10, 20] dB\n");
    printf("- phase margin should be greater than 45°\n");
    printf("Gain margin : %f\nPhase margin : %f\n",gain_margin, phase_margin);
    printf("\n");
    printf("The roots of Hboc denominator are :");
    disp(racines);
    printf("\n");
endfunction

